// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#include "msg.h"
#include "vars.h"
#include "misc.h"
#include "furnace.h"

#define COOL_FAN_TIME (3 * 60)
#define MIN_AC_OFF_TIME (30 * 60)
#define MIN_AUTOSWITCH_GAP 1.0

#if TEST
#undef COOL_FAN_TIME
#undef MIN_AC_OFF_TIME
#define COOL_FAN_TIME 3
#define MIN_AC_OFF_TIME 10
#endif

typedef enum {
  zs_off,
  zs_heat_on,
  zs_heat_off,
  zs_cool_on,
  zs_cool_fan,
  zs_cool_off,
} ZoneState;

struct Zone {
  int z;

  Var max_heat;
  Var min_heat;
  Var max_cool;
  Var min_cool;
  Var max_humidity;
  Var min_humidity;

  double temp;
  double humidity;
  int humidifying;

  ZoneState state;
  void set_state(ZoneState zs);
  time_t since;

  Zone(int z);
};

static char *
ts(void)
{
  static char buf[40];
  time_t now;
  struct tm *t;
  time(&now);
  t = localtime(&now);
  sprintf(buf, "%02d:%02d", t->tm_hour, t->tm_min);
  return buf;
}

//-----------------------------------------------------------------------------

int alarm_time = 2000;

static double saved_temps[4];

static const char *zsnames[] = {
  "off",
  "heat-on", "heat-fan", "heat-off",
  "cool-on", "cool-fan", "cool-off"
};

Var heating("heating-mode");
Var cooling("cooling-mode");
Var last_ac_on("last-ac-on");
Var autocool("auto-cool-temp");
Var autoheat("auto-heat-temp");

FurnaceData furnace, old_furnace;

double outside_temp, outside_humidity;

Zone zone[3] = {
  Zone(0),
  Zone(1),
  Zone(2)
};

//-----------------------------------------------------------------------------

Zone::Zone(int Pz)
: max_heat(Pz, "max-heat"),
  min_heat(Pz, "min-heat"),
  max_cool(Pz, "max-cool"),
  min_cool(Pz, "min-cool"),
  max_humidity(Pz, "max-humidity"),
  min_humidity(Pz, "min-humidity")
{
  z = Pz;
  since = 0;
  temp = 70;
  humidity = 40;
}

void
Zone::set_state (ZoneState zs)
{
  if (zs == state)
    return;
  time(&since);
  //  printf("%s zone %d: %s -> %s\n", ts(), z, zsnames[state], zsnames[zs]);
  saved_temps[0] = -1;
  state = zs;
  if (zs != zs_heat_on)
    humidifying = 0;
}

//-----------------------------------------------------------------------------

static void
var_range (Var &min, Var &max, double minv, double maxv)
{
  if (max > maxv)
    max.set(maxv);
  if (min < minv)
    min.set(minv);
  if (min >= max)
    min.set(max - 1);
}

void
sanity_check_variables (void)
{
  int z;

  if (heating != 0 && heating != 1)
    heating.set(0);
  if (cooling != 0 && cooling != 1)
    cooling.set(0);
  if (heating && cooling)
    cooling.set(0);

  if (last_ac_on == 0 && last_ac_on.id)
    {
      time_t now;
      time(&now);
      /* printf("%s last_ac_on: assume was on before crash\n", ts()); */
      last_ac_on.set(now);
    }

  if (autoheat < 50)
    autoheat.set(50);
  if (autocool > 90)
    autocool.set(90);

  for (z=0; z<3; z++)
    {
      var_range (zone[z].min_heat, zone[z].max_heat, 55, 85);
      var_range (zone[z].min_cool, zone[z].max_cool, 55, 85);
      var_range (zone[z].min_humidity, zone[z].max_humidity, 0, 60);

      if (autoheat > zone[z].min_cool - MIN_AUTOSWITCH_GAP)
	autoheat.set(zone[z].min_cool - MIN_AUTOSWITCH_GAP);
      if (autocool < zone[z].max_heat + MIN_AUTOSWITCH_GAP)
	autocool.set(zone[z].max_heat + MIN_AUTOSWITCH_GAP);
    }
}

void
note_ac_last_on (void)
{
  static int old_cool = -1;

  if (old_cool == furnace.cool)
    return;
  old_cool = furnace.cool;
  if (!furnace.cool && last_ac_on == 0)
    {
      time_t t;
      time(&t);
      //printf("%s last_ac_on: setting\n", ts());
      last_ac_on.set(t);
    }
}

void
send_furnace ()
{
  static int saved_mode = -1;
  note_ac_last_on ();

  if (furnace.cool)
    furnace.fan = 1;
  if (furnace.heat)
    furnace.fan = 0;
  if (!furnace.heat)
    furnace.humidify = 0;
  if (!furnace.zone[0] && !furnace.zone[1] && !furnace.zone[2])
    furnace.zone[0] = furnace.zone[1] = furnace.zone[2] = 1;

  if (memcmp (&old_furnace, &furnace, sizeof(FurnaceData)))
    {
#if 0
      printf("%s Furnace: %s %s %s %s : %s %s %s %s\n", ts(),
	     furnace.heat ? "H" : "-",
	     furnace.cool ? "C" : "-",
	     furnace.fan ? "F" : "-",
	     furnace.humidify ? "W" : "-",
	     furnace.zone[0] ? "0" : "-",
	     furnace.zone[1] ? "1" : "-",
	     furnace.zone[2] ? "2" : "-",
	     furnace.zone[3] ? "3" : "-");
#endif
      alarm_time = 2000;
      old_furnace = furnace;
    }
#if 0
  else if (alarm_time < 10 * 1000)
    alarm_time += 1 * 1000;
#endif

  if (saved_mode != (int)(heating*2+cooling))
    {
      saved_mode = (int)(heating*2+cooling);
#if 0
      if (heating)
	printf("%s Heating mode\n", ts());
      else if (cooling)
	printf("%s Cooling mode\n", ts());
      else
	printf("%s Do-nothing mode\n", ts());
#endif
    }

  Message f(&furnace, sizeof(furnace), MT_set_furnace);
  f.send();
}

int
do_overrides (void)
{
  int z;

  for (z=0; z<3; z++)
    if (zone[z].temp < 50)
      {
	saved_temps[0] = -1;
	printf("%s heat override: ", ts());
	furnace.heat = 1;
	furnace.fan = 1;
	furnace.humidify = 0;
	furnace.cool = 0;
	for (z=0; z<3; z++)
	  if (zone[z].temp < 50)
	    {
	      printf(" %d (%f)", z, zone[z].temp);
	      furnace.zone[z] = 1;
	      zone[z].set_state(zs_heat_on);
	    }
	  else
	    furnace.zone[z] = 0;
	printf("\n");
	send_furnace();
	if (cooling)
	  cooling.set(0);
	if (!heating)
	  heating.set(1);
	return 1;
      }

  return 0;
}

void
do_heating (void)
{
  int z;
  time_t now;

  // if it gets really warm outside, switch to do-nothing mode.
  if (outside_temp > autocool)
    {
      heating.set(0);
      return;
    }

  time(&now);

  // old_furnace has previous values.
  memset(&furnace, 0, sizeof(furnace));

  for (z=0; z<3; z++)
    {
      switch (zone[z].state)
	{
	case zs_heat_on:
	  if (zone[z].temp > zone[z].max_heat)
	    zone[z].set_state(zs_heat_off);
	  else
	    {
	      furnace.heat = 1;
	      furnace.zone[z] = 1;
	    }
	  break;

	case zs_heat_off:
	  // if it's warmer outside than in, don't run the furnace.
	  if (zone[z].max_heat < outside_temp)
	    break;
	  if (zone[z].temp < zone[z].min_heat
	      || (old_furnace.heat
		  && zone[z].temp < (zone[z].min_heat+zone[z].max_heat)/2))
	    {
	      zone[z].set_state(zs_heat_on);
	      furnace.heat = 1;
	      furnace.zone[z] = 1;
	    }
	  /* else leave it off */
	  break;

	default:
	  zone[z].set_state(zs_heat_off);
	  break;
	}
    }

  if (furnace.heat)
    {
      for (z=0; z<3; z++)
	{
	  if (furnace.zone[z]
	      && (zone[z].humidity < zone[z].min_humidity
		  || (zone[z].humidifying
		      && zone[z].humidity < zone[z].max_humidity)))
	    {
	      furnace.humidify = 1;
	      zone[z].humidifying = 1;
	    }
	  else
	    zone[z].humidifying = 0;
	}
    }

  send_furnace();
}

void
do_cooling (void)
{
  int z;
  time_t now;
  int ok_to_cool;
  int n_cools = 0;
  int n_old_cools = 0;

  // If we see cold enough outside temps, switch to do-nothing mode.
  if (outside_temp < autoheat)
    {
      cooling.set(0);
      return;
    }

  time(&now);

  // old_furnace has previous values.
  memset(&furnace, 0, sizeof(furnace));

  ok_to_cool = (old_furnace.cool
		|| (now - last_ac_on > MIN_AC_OFF_TIME));

  for (z=0; z<3; z++)
    {
      switch (zone[z].state)
	{
	case zs_cool_on:
	  if (zone[z].temp < zone[z].min_cool)
	    {
	      zone[z].set_state(zs_cool_fan);
	      furnace.fan = 1;
	      furnace.zone[z] = 1;
	    }
	  else
	    {
	      furnace.cool = 1;
	      furnace.zone[z] = 1;
	      n_cools ++;
	      n_old_cools ++;
	    }
	  break;

	case zs_cool_fan:
	  if (now - zone[z].since > COOL_FAN_TIME)
	    {
	      zone[z].set_state(zs_cool_off);
	    }
	  else
	    {
	      furnace.fan = 1;
	    }
	  break;

	case zs_cool_off:
	  // if it's cooler outside than in, don't run the AC.
	  if (zone[z].min_cool > outside_temp)
	    break;
	  if (ok_to_cool
	      && (zone[z].temp > zone[z].max_cool
		  || (old_furnace.cool
		      && zone[z].temp > (zone[z].min_cool+zone[z].max_cool)/2)))
	    {
	      zone[z].set_state(zs_cool_on);
	      furnace.cool = 1;
	      furnace.zone[z] = 1;
	      n_cools ++;
	    }
	  /* else leave it off */
	  break;

	default:
	  zone[z].set_state(zs_cool_off);
	  break;
	}
    }

  if (n_cools > 1)
    {
      if (n_old_cools)
	{
	  for (z=0; z<3; z++)
	    if (furnace.zone[z] && !old_furnace.zone[z])
	      {
		furnace.zone[z] = 0;
		zone[z].state = zs_cool_off;
	      }
	}
      else
	{
	  int best_z = -1;
	  double best_t = 0;
	  for (z=0; z<3; z++)
	    if (furnace.zone[z])
	      if (best_z == -1
		  || zone[z].temp - zone[z].min_cool > best_t)
		{
		  best_z = z;
		  best_t = zone[z].temp - zone[z].min_cool;
		}
	  for (z=0; z<3; z++)
	    if (furnace.zone[z] && z != best_z)
	      {
		furnace.zone[z] = 0;
		zone[z].state = zs_cool_off;
	      }
	}
    }

  if (furnace.cool)
    {
      if (last_ac_on)
	last_ac_on.set(0);

      for (z=0; z<3; z++)
	{
	  if (zone[z].state == zs_cool_fan)
	    {
	      zone[z].set_state(zs_cool_off);
	      furnace.zone[z] = 0;
	    }
	}
    }

  send_furnace();
}

void
do_nothing (void)
{
  double ave;

  ave = (zone[0].temp + zone[1].temp + zone[2].temp) / 3;
  if (ave > autocool)
    {
      cooling.set(1);
      return;
    }
  if (ave < autoheat)
    {
      heating.set(1);
      return;
    }

  furnace.heat = 0;
  furnace.fan = 0;
  furnace.cool = 0;
  furnace.humidify = 0;
  furnace.zone[0] = 1;
  furnace.zone[1] = 1;
  furnace.zone[2] = 1;
  furnace.zone[3] = 0;
  send_furnace ();
}

void
arbitrate (void)
{
  int z;

#if 0
  for (z=0; z<4; z++)
    if (saved_temps[z] != zone[z].temp)
      {
	printf(" \narbitrating!  ");
	for (z=0; z<4; z++)
	  {
	    printf("  z%d: %4.1fF %d%%", z, (double)zone[z].temp, (int)zone[z].humidity);
	    saved_temps[z] = zone[z].temp;
	  }
	printf("\n");
	break;
      }
  if (z == 4)
    {
      static int twiddle = 0;
      twiddle = (twiddle+1)%4;
      printf("%c\r", "|/-\\"[twiddle]);
      fflush(stdout);
    }
#endif

  if (do_overrides ())
    return;

  if (heating)
    do_heating();
  else if (cooling)
    do_cooling();
  else
    do_nothing();
}

void
set_reasonable_defaults (void)
{
  int z;
  for (z=0; z<3; z++)
    {
      if (zone[z].max_heat == 0)
	{
	  zone[z].max_heat.set(70);
	  zone[z].min_heat.set(66);
	}
      if (zone[z].max_cool == 0)
	{
	  zone[z].max_cool.set(76);
	  zone[z].min_cool.set(72);
	}
      if (zone[z].max_humidity == 0)
	{
	  zone[z].max_humidity.set(35);
	  zone[z].min_humidity.set(30);
	}
      zone[z].temp = (zone[z].min_heat + zone[z].max_cool) / 2;
    }
  sanity_check_variables ();
  memset(&furnace, 0, sizeof(furnace));
  send_furnace();
}

static void
exit_proc(int)
{
  memset(&furnace, 0, sizeof(furnace));
  send_furnace();
  exit(0);
}

int
main(int argc, char **argv)
{
  Message *m;

  signal(SIGTERM, exit_proc);
  signal(SIGINT, exit_proc);

  Message::set_my_name("arbiter");
  Message::init();

  Var::wait_for_all_ids();

  set_reasonable_defaults();

  Message::set_alarm(alarm_time);
  while (1)
    {
      m = Message::next();

      if (!m)
	{
	  arbitrate();
	  Message::set_alarm(alarm_time);
	  continue;
	}

      switch (m->type)
	{
	case MT_set_furnace:
	  {
	    FurnaceData *fd = (FurnaceData *)m->data;
	    furnace = *(FurnaceData *)m->data;
	    note_ac_last_on();
	    break;
	  }

	case MT_climate:
	  {
	    ClimateData *cd = (ClimateData *)m->data;
	    if (cd->zone == 3)
	      {
		outside_temp = cd->temperature;
		outside_humidity = cd->humidity;
	      }
	    else if (cd->zone <= 2)
	      {
		zone[cd->zone].temp = cd->temperature;
		zone[cd->zone].humidity = cd->humidity;
	      }
	    break;
	  }

	case MT_set_var_value:
	  sanity_check_variables ();
	  break;

	}
    }
}
