// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include "msg.h"
#include "msghdr.h"
#include "vars.h"

#define MT_NAMELIST
#include "msgtypes.h"

static int mux_fd = -1;
static int my_id = -1;
static int snooping = 0;
static const char *hostname = 0;

Message::Message(int Plength, MsgType mtype)
{
  source = my_id;
  dest = DEST_ALL;
  type = mtype;
  length = Plength;
  data = new unsigned char [length];
  my_data = 1;
}

Message::Message(const void *pdata, int plength, MsgType mtype)
{
  source = my_id;
  dest = DEST_ALL;
  type = mtype;
  data = (unsigned char *)pdata;
  length = plength;
  my_data = 0;
}

Message::~Message()
{
  if (my_data)
    delete data;
}

static const char *my_name = "unknown";

static int max_name = 0;
static char **name_list = 0;

int
Message::mux_socket (void)
{
  return mux_fd;
}

static void
set_name_for (const char *name, int id, int length)
{
  if (max_name <= id)
    {
      int newmax = id + 10;
      if (name_list)
	name_list = (char **)realloc(name_list, newmax * sizeof(char *));
      else
	name_list = (char **)malloc(newmax * sizeof(char *));
      while (max_name < newmax)
	name_list[max_name++] = 0;
    }
  if (name_list[id])
    delete name_list[id];
  name_list[id] = (char *)malloc(length);
  memcpy (name_list[id], name, length);
}

void
Message::set_my_name (const char *name)
{
  my_name = name;
}

const char *
Message::get_name_for (int id)
{
  static char buf[4][10];
  static int bufi = 0;
  if (id > 0 && id < max_name && name_list[id])
    return name_list[id];
  if (id == DEST_ALL)
    return "all";
  if (id == DEST_MUX)
    return "mux";
  bufi = (bufi + 1) % 4;
  sprintf(buf[bufi], "[%d]", id);
  return buf[bufi];
}

void
Message::set_snoop(int on)
{
  static char mode[1];
  snooping = on;
  if (my_id != -1)
    {
      Message m(mode, 1, MT_snoop);
      m.dest = DEST_MUX;
      mode[0] = on;
      m.send();
    }
}

void
Message::set_host(const char *Phostname)
{
  hostname = Phostname;
  printf("msg: host set to `%s'\n", hostname);
}

static int
muxwrite(void *pdata, int len)
{
  char *data = (char *)pdata;
  while (1)
    {
      int n = write(mux_fd, data, len);
      if (n == len)
	return 1;
      if (n <= 0)
	return -1;
      data += n;
      len -= n;
    }
}

void
Message::send()
{
  MessageHeader h;

  if (mux_fd == -1)
    init();

  h.length = length;
  h.dest_id = dest;
  h.source_id = source;
  h.type = type;
  muxwrite(&h, sizeof(h));
  muxwrite(data, length);
}

void
Message::init(void)
{
  struct sockaddr_in addr;
  int one = 1;

  if (!my_name)
    {
      printf("Must set_my_name before init\n");
      abort();
    }

  mux_fd = socket (PF_INET, SOCK_STREAM, 0);
  setsockopt (mux_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

  addr.sin_family = AF_INET;
  addr.sin_port = htons(MUX_PORT);
  if (hostname)
    {
      struct hostent *he;
      he = gethostbyname (hostname);
      if (!he)
	{
	  fprintf(stderr, "msg: unknown host `%s'\n", hostname);
	  exit(1);
	}
      memcpy (&addr.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
    }
  else
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

  for (int i=0; ; i++)
    {
      if (i)
	{
	  unsigned char *ip = (unsigned char *)&addr.sin_addr.s_addr;
	  printf("attempting to connect to %d.%d.%d.%d ...\n",
		 ip[0], ip[1], ip[2], ip[3]);
	}
      if (connect (mux_fd, (sockaddr *)&addr, sizeof(addr)) == 0)
	{
	  Message *m = next();
	  if (m)
	    {
	      if (m->type != MT_my_name)
		printf("hmmm... mux didn't send us an MT_my_name intro message.\n");
	      else
		{
		  my_id = m->dest;
		  set_name_for (my_name, my_id, strlen(my_name));
		  break;
		}
	    }
	}
      if (i == 5)
	{
	  perror("connect");
	  exit(1);
	}
      sleep (1);
    }
  setsockopt (mux_fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
  setsockopt (mux_fd, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));

  if (snooping)
    set_snoop(1);

  Message name(strlen(my_name)+1, MT_ask_name);
  name.dest = DEST_ALL;
  strcpy ((char *)name.data, my_name);
  name.send();
}

static int
muxread (void *pdata, int len)
{
  char *data = (char *)pdata;
  while (1)
    {
      int n = read(mux_fd, data, len);
      if (n == len)
	return 1;
      if (n <= 0)
	return -1;
      data += n;
      len -= n;
    }
}

static int alarm_set = 0;
static struct timeval alarm_time;

void
Message::set_alarm(int msec_from_now)
{
  struct timeval now;
  if (msec_from_now == -1)
    {
      alarm_set = 0;
      return;
    }
  gettimeofday(&now, 0);
  alarm_time.tv_usec = now.tv_usec + msec_from_now * 1000;
  alarm_time.tv_sec = now.tv_sec + (alarm_time.tv_usec / 1000000);
  alarm_time.tv_usec %= 1000000;
  alarm_set = 1;
}

Message *
Message::next(int waitms)
{
  static Message *rv = 0;
  static MessageHeader mh;
  static int max_length = 0;

  if (mux_fd == -1)
    init();

  if (alarm_set)
    {
      struct timeval now;
      gettimeofday(&now, 0);
      int dsec = (int)(alarm_time.tv_sec - now.tv_sec);
      int dusec = (int)(alarm_time.tv_usec - now.tv_usec);
      int a_ms = dsec * 1000 + dusec / 1000;
      if (a_ms <= 0)
	{
	  alarm_set = 0;
	  return 0;
	}
      if (waitms == -1 || waitms >= a_ms)
	waitms = a_ms;
    }

  if (waitms != -1)
    {
      fd_set rfd;
      struct timeval tv;
      FD_ZERO (&rfd);
      FD_SET (mux_fd, &rfd);
      tv.tv_sec = waitms/1000;
      tv.tv_usec = (waitms%1000) * 1000;
      if (select (mux_fd+1, &rfd, 0, 0, &tv) == 0)
	{
	  alarm_set = 0;
	  return 0;
	}
    }

  if (muxread(&mh, sizeof(mh)) < 0)
    {
      init();
      return next(waitms);
    }

  if (rv == 0)
    {
      rv = new Message(mh.length, MT_NOP);
      rv->dest = mh.dest_id;
      max_length = mh.length;
    }
  else if (max_length < mh.length)
    {
      delete rv->data;
      rv->data = new unsigned char[mh.length];
      max_length = mh.length;
    }

  rv->source = mh.source_id;
  rv->dest = mh.dest_id;
  rv->type = (MsgType)mh.type;
  rv->length = mh.length;

  if (muxread(rv->data, rv->length) < 0)
    {
      init();
      return next(waitms);
    }

  if (rv->source != my_id)
    switch (rv->type)
      {
      case MT_my_name:
      case MT_ask_name:
	set_name_for ((char *)rv->data, rv->source, rv->length);
	if (rv->type == MT_ask_name)
	  {
	    Message name(strlen(my_name)+1, MT_my_name);
	    name.dest = rv->source;
	    strcpy ((char *)name.data, my_name);
	    name.send();
	  }
	break;

      case MT_var_id:
      case MT_set_var_value:
	Var::handle_message(rv);
	break;
      }

  return rv;
}
