// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "msg.h"
#include "vars.h"

// There are three files that we store variables in.
//
// vardb.values = array of VarValueType
// vardb.nameofs = array of int (offset into vardb.namestr).  Zero
//   means unused.  The first entry is the number of vars.  Index into
//   this is the var's ID.
// vardb.namestr = block of ASCIIZ strings. first byte is zero; unused.
//
// Values in files are always in the same order as each other;
// including the namestr block.

struct Mapping {
  const char *filename;
  int fd;
  void *addr;
  int length;
  Mapping(const char *filename);
  ~Mapping();
  void map(void);
  void need(int length);
  void sync();
};

Mapping::Mapping(const char *fn)
{
  filename = fn;
  fd = -1;
  addr = 0;
  length = 0;
}

Mapping::~Mapping()
{
  munmap(addr, length);
  close(fd);
}

static int pagesize = 0;

#define ROUND_UP(v,r) ((int)(((v) + (r) - 1)/(r)) * (r))

void
Mapping::map (void)
{
  struct stat st;

  if (pagesize == 0)
    pagesize = getpagesize();

  fd = open(filename, O_RDWR|O_CREAT, 0666);
  if (fd < 0)
    perror(filename);
  fstat(fd, &st);
  length = st.st_size;
  if (length == 0)
    length = 1;
  if (length % pagesize)
    {
      length = ROUND_UP(length, pagesize);
      lseek (fd, length-1, SEEK_SET);
      write(fd, "\0", 1);
    }
  addr = mmap (0, length, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);

  if (st.st_size == 0)
    // New file
    *(int *)addr = 0;
}

void
Mapping::need (int l)
{
  if (l <= length)
    return;
  munmap (addr, length);
  length = ROUND_UP(l, pagesize);
  lseek (fd, length-1, SEEK_SET);
  write(fd, "\0", 1);
  addr = mmap (0, length, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
}

void
Mapping::sync()
{
  msync (addr, length, MS_SYNC | MS_INVALIDATE);
}

Mapping values_map("vardb.values");
Mapping nameidx_map("vardb.nameidx");
Mapping namestr_map("vardb.namestr");

#define values ((VarValueType *)values_map.addr)
#define nameidx ((int *)nameidx_map.addr)
#define namestr ((char *)namestr_map.addr)
#define num_vars (nameidx[0])

#define name(ID) (namestr + nameidx[ID])

static void
do_ask_var_id (Message *m)
{
  int id;
  char *name = (char *)m->data;


  for (id=1; id<num_vars+1; id++)
    if (strcmp (name(id), name) == 0)
      {
	//	printf("varserver: ask id for %s = %d = %g\n", name, id, values[id]);
	break;
      }

  if (id == num_vars+1)
    {
      id = num_vars+1;
      //      printf("varserver: ask id for %s: new %d\n", name, id);
      num_vars++;

      values_map.need ((id+1)*sizeof(values[0]));
      values[id] = 0;

      nameidx_map.need ((id+1)*sizeof(nameidx[0]));
      nameidx[id] = nameidx[id-1] + strlen(name(id-1)) + 1;

      namestr_map.need(nameidx[id]+strlen(name)+1);
      strcpy(name(id), name);

      values_map.sync();
      nameidx_map.sync();
      namestr_map.sync();
    }

  Message reply(sizeof(VarIdData) + strlen(name), MT_var_id);
  reply.dest = m->source;
  ((VarIdData *)reply.data)->id = id;
  ((VarIdData *)reply.data)->value = values[id];
  strcpy (((VarIdData *)reply.data)->name, name);
  reply.send();
}

static void
do_set_var_value (Message *m)
{
  SetVarValueData *svd = (SetVarValueData *)m->data;
  if (svd->id > 0 && svd->id <= num_vars)
    if (svd->value != values[svd->id])
      {
	//	printf("varserver: set var %d %s = %g\n", svd->id, name(svd->id), svd->value);
	values[svd->id] = svd->value;
	if (m->dest != DEST_ALL)
	  {
	    m->dest = DEST_ALL;
	    m->send();
	  }
      }
}

static void
do_ask_var_value (Message *m)
{
  int id = *(int *)m->data;
  //  printf("varserver: ask var value %d (%s) = %g\n", id, name(id), values[id]);
  if (id > 0 && id <= num_vars)
    {
      SetVarValueData svd;
      svd.id = id;
      svd.value = values[id];
      Message reply(&svd, sizeof(svd), MT_set_var_value);
      reply.dest = m->source;
      reply.send();
    }
}

void
sigexit(int v)
{
  exit(0);
}

int
main(int argc, char **argv)
{
  Message *m;
  char *progdir = strdup(argv[0]);
  if (strrchr(progdir, '/'))
    {
      *strrchr(progdir, '/') = 0;
      chdir(progdir);
    }

  signal(SIGINT, sigexit);
  signal(SIGHUP, sigexit);
  signal(SIGTERM, sigexit);

  values_map.map();
  nameidx_map.map();
  namestr_map.map();

  Message::set_my_name("varserver");
  Message::init();

  char *name;
  int id;
  VarValueType value;

  while ((m = Message::next()) != NULL)
    switch (m->type)
      {
      case MT_ask_var_id:
	do_ask_var_id(m);
	break;

      case MT_set_var_value:
	do_set_var_value(m);
	break;

      case MT_ask_var_value:
	do_ask_var_value(m);
	break;

      }
}
